from pydantic import BaseModel, Field, field_validator
import re


class Region(BaseModel):
    __namePattern = re.compile("^[a-z0-9]+( [a-z0-9]+)?( [a-z0-9]+)?$", re.IGNORECASE)
    __locPattern = re.compile(
        r"^[a-z0-9]+( [a-z0-9]+)?( [a-z0-9]+)?\s*\(\s*\d+\s*\,\s*\d+\s*\)$",
        re.IGNORECASE,
    )
    name: str = Field(min_length=3, max_length=25, pattern=__namePattern)
    location: tuple[int, int]

    def __init__(self, *args, **kwargs) -> None:
        # constructor by str <Region name (0, 0)>
        if len(args):
            nm, xy = args[0].split("(")
            nm = nm.strip()
            x, y = xy.replace(")", "").strip().split(",")
            super().__init__(name=nm, location=(x, y))
        else:
            super().__init__(**kwargs)

    def __str__(self) -> str:
        return f"{self.name} {self.location}"

    @field_validator("location")
    def location_validator(cls, location: tuple[int, int]):
        if location[0] < 0 or location[1] < 0:
            raise ValueError("Region.location must contains positive values")
        if location[0] % 256 or location[1] % 256:
            raise ValueError("Region.location must contains in 256x256 grid values")
        return location



from pydantic import BaseModel, Field, field_validator, AliasChoices

from .avatar import Avatar


class Touch(BaseModel):
    prim: int = Field(
        default=0, ge=0, le=255, validation_alias=AliasChoices("prim", "prim_num")
    )
    face: int = Field(
        default=0, ge=0, le=9, validation_alias=AliasChoices("face", "side")
    )
    avatar: Avatar
    startST: tuple[float, float]
    endST: tuple[float, float]
    startUV: tuple[float, float]
    endUV: tuple[float, float]

    @field_validator("startST")
    def startST_validator(cls, coords: tuple[float, float]):
        if coords[0] < 0 or coords[1] < 0:
            raise ValueError("Touch.startST must contains positive values")
        if coords[0] > 1 or coords[1] > 1:
            raise ValueError("Touch.startST must contains values less than 1")
        return coords

    @field_validator("endST")
    def endST_validator(cls, coords: tuple[float, float]):
        if coords[0] < 0 or coords[1] < 0:
            raise ValueError("Touch.endST must contains positive values")
        if coords[0] > 1 or coords[1] > 1:
            raise ValueError("Touch.endST must contains values less than 1")
        return coords

    @field_validator("startUV")
    def startUV_validator(cls, coords: tuple[float, float]):
        if coords[0] < 0 or coords[1] < 0:
            raise ValueError("Touch.startUV must contains positive values")
        if coords[0] > 1 or coords[1] > 1:
            raise ValueError("Touch.startUV must contains values less than 1")
        return coords

    @field_validator("endUV")
    def endUV_validator(cls, coords: tuple[float, float]):
        if coords[0] < 0 or coords[1] < 0:
            raise ValueError("Touch.endUV must contains positive values")
        if coords[0] > 1 or coords[1] > 1:
            raise ValueError("Touch.endUV must contains values less than 1")
        return coords



from pydantic import BaseModel, Field, AliasChoices
from datetime import datetime
from uuid import UUID


class PrimInfo(BaseModel):
    # id calculates on client side
    id: UUID
    creatorId: UUID
    name: str = Field(pattern=r"^[\x20-\x7b\x7d-\x7e]{1,63}$")
    description: str = Field(pattern=r"^[\x20-\x7b\x7d-\x7e]{0,127}$")
    faces: int = Field(
        default=1,
        gt=0,
        le=9,
        validation_alias=AliasChoices("faces", "faces_num", "sides", "sides_num"),
    )
    createdAt: datetime



from pydantic import BaseModel, Field, AliasChoices, field_validator
from uuid import UUID


from lslgwlib.enums import ChatChannel


# https://wiki.secondlife.com/wiki/Listen
class ChatMessage(BaseModel):
    """Message from chat model

    Event: listen(integer channel, string name, key id, string message)
    """

    channel: int = Field(default=ChatChannel.PUBLIC, ge=-0x80000000, le=0x7FFFFFFF)
    name: str = Field(pattern=r"^[\x20-\x7b\x7d-\x7e]{1,63}$")
    id: UUID = Field(default=UUID(int=0))
    message: str = Field(
        min_length=1, max_length=1024, validation_alias=AliasChoices("msg", "message")
    )

    @field_validator("message")
    def msg_validator(cls, msg: str) -> str:
        if len(msg.encode("UTF-8")) > 1024:
            raise ValueError("max length of message is 1024 bytes")
        return msg



from enum import IntEnum


# https://wiki.secondlife.com/wiki/Category:LSL_Integer/link
class LinkNum(IntEnum):
    UNLINKED = 0  #      only one prim object, and there are no avatars seated upon it
    ROOT = 1  #          sends to the root prim in a multi-prim linked set
    SET = -1  #          sends to all prims
    ALL_OTHERS = -2  #   sends to all other prims
    ALL_CHILDREN = -3  # sends to all children, (everything but the root)
    THIS = -4  #         sends to the prim the script is in



from pydantic import BaseModel, Field
from uuid import UUID

from .avatar import Avatar
from .region import Region


class HTTPData(BaseModel):
    """HTTP request model"""

    # info from X-SecondLife-* headers:
    objectKey: UUID
    objectName: str = Field(pattern=r"^[\x20-\x7b\x7d-\x7e]{1,63}$")
    owner: Avatar
    position: tuple[float, float, float]
    rotation: tuple[float, float, float, float]
    velocity: tuple[float, float, float]
    region: Region
    production: bool

    # response (different by API methods)
    data: (
        BaseModel
        | list[BaseModel | int | float | str | UUID]
        | int
        | float
        | str
        | UUID
        | None
    )
    # other http headers
    headers: dict[str, str]



from enum import IntEnum


# https://wiki.secondlife.com/wiki/Listen
# https://wiki.secondlife.com/wiki/LlSay
class ChatChannel(IntEnum):
    PUBLIC = 0  #         Chat channel that broadcasts to all nearby users
    DEBUG = 0x7FFFFFFF  # Chat channel reserved for script debugging and error messages



from pydantic import BaseModel, Field
from uuid import UUID
import re


class Avatar(BaseModel):
    """Avatar model"""

    __namePattern = re.compile("^[a-z0-9]{1,31}$", re.IGNORECASE)
    __nameStrPattern = re.compile(
        r"^[a-z0-9]{1,31}((\ |\.)[a-z0-9]{1,31})?$", re.IGNORECASE
    )
    id: UUID
    firstName: str = Field(pattern=__namePattern)
    lastName: str | None = Field(default=None, pattern=__namePattern)

    def __init__(self, *args, **kwargs) -> None:
        # constructor by id:uuid, fullname:str
        if len(args) == 2:
            id, name = args
            try:
                fn, ln = name.replace(".", " ").split()
            except ValueError:
                fn = name
                ln = None
            if ln and ln.lower() == "resident":
                ln = None
            super().__init__(id=id, firstName=fn, lastName=ln)
        else:
            super().__init__(**kwargs)

    def __eq__(self, o) -> bool:
        return self.modernName() == o.modernName() and self.id == o.id

    def legacyName(self) -> str:
        """Get name in legacy format <Firstname Lastname>"""
        ln = self.lastName
        if not ln:
            ln = "Resident"
        return f"{self.firstName} {ln}"

    def modernName(self) -> str:
        """Get name in modern format <username.resident>"""
        ln = self.lastName
        if ln:
            ln = ln.lower()
        else:
            ln = "resident"
        return f"{self.firstName.lower()}.{ln}"



from pydantic import BaseModel, Field, model_validator
from typing import Pattern, Sequence
from datetime import datetime
from uuid import UUID
import re

from lslgwlib.enums import InvetoryType
from .permissions import Permissions


class InvetoryItem(BaseModel):
    """Invetory item model"""

    id: UUID
    type: InvetoryType
    name: str = Field(pattern=r"^[\x20-\x7b\x7d-\x7e]{1,63}$")
    description: str = Field(pattern=r"^[\x20-\x7b\x7d-\x7e]{0,127}$")
    creatorId: UUID
    permissions: Permissions
    acquireTime: datetime


class Invetory(BaseModel):
    """LSL linkset inventory model

    Fields:
    items -    list of inventory items (InvetoryItem model)
    filtered - indicates if inventory loaded by items type (not full representation)
    """

    items: list[InvetoryItem] = Field(max_length=10000)
    filtered: InvetoryType = Field(default=InvetoryType.ANY)

    @model_validator(mode="after")
    def check_filtered(self):
        if self.filtered != InvetoryType.ANY:
            if any(x.type != self.filtered for x in self.items):
                raise ValueError(
                    "Filtered Inventory must contains only items with same type"
                )
        return self

    def byName(self, name: str) -> InvetoryItem | None:
        """Get item by its name

        Arguments:
        name - string
        """
        for x in self.items:
            if x.name == name:
                return x
        return None

    def byNamePattern(self, pattern: str | Pattern) -> list[InvetoryItem]:
        """Get items by name part or regex

        Arguments:
        pattern - string part of name or regex
        """
        items = list()
        for x in self.items:
            if isinstance(pattern, Pattern) and re.match(pattern, x.name):
                items.append(x)
            if isinstance(pattern, str) and pattern in x.name:
                items.append(x)
        return items

    def byType(self, types: Sequence[InvetoryType]) -> list[InvetoryItem]:
        """Get items by types

        Arguments:
        types - sequence of types
        """
        if InvetoryType.ANY in types:
            return self.items
        for bytype in types:
            if self.filtered != InvetoryType.ANY and self.filtered != bytype:
                raise ValueError(
                    f"This Inventory instance loaded without {repr(bytype)}"
                )
        return list(filter(lambda x: (x.type in types), self.items))

    def names(self) -> list[str]:
        return list(map(lambda x: x.name, self.items))

    @property
    def textures(self) -> list[InvetoryItem]:
        return self.byType([InvetoryType.TEXTURE])

    @property
    def sounds(self) -> list[InvetoryItem]:
        return self.byType([InvetoryType.SOUND])

    @property
    def landmarks(self) -> list[InvetoryItem]:
        return self.byType([InvetoryType.LANDMARK])

    @property
    def clothings(self) -> list[InvetoryItem]:
        return self.byType([InvetoryType.CLOTHING])

    @property
    def objects(self) -> list[InvetoryItem]:
        return self.byType([InvetoryType.OBJECT])

    @property
    def notecards(self) -> list[InvetoryItem]:
        return self.byType([InvetoryType.NOTECARD])

    @property
    def scripts(self) -> list[InvetoryItem]:
        return self.byType([InvetoryType.SCRIPT])

    @property
    def bodyparts(self) -> list[InvetoryItem]:
        return self.byType([InvetoryType.BODYPART])

    @property
    def animations(self) -> list[InvetoryItem]:
        return self.byType([InvetoryType.ANIMATION])

    @property
    def gestures(self) -> list[InvetoryItem]:
        return self.byType([InvetoryType.GESTURE])

    @property
    def settings(self) -> list[InvetoryItem]:
        return self.byType([InvetoryType.SETTING])

    @property
    def materials(self) -> list[InvetoryItem]:
        return self.byType([InvetoryType.MATERIAL])



from pydantic import BaseModel, Field, AliasChoices, field_validator
from uuid import UUID


from lslgwlib.enums import LinkNum


# https://wiki.secondlife.com/wiki/Link_message
# https://wiki.secondlife.com/wiki/LlMessageLinked
class LinkMessage(BaseModel):
    """Link message

    Event: link_message(integer sender_num, integer num, string str, key id)
    Function: llMessageLinked(integer link, integer num, string str, key id)
    """

    prim: int = Field(
        default=LinkNum.THIS,
        ge=LinkNum.THIS,
        le=255,
        validation_alias=AliasChoices(
            "prim",
            "prim_num",
            "sender",
            "sender_num",
            "link",
            "source",
            "from",
            "target",
            "to",
        ),
    )
    num: int = Field(default=0, ge=-0x80000000, le=0x7FFFFFFF)
    string: str = Field(default="", validation_alias=AliasChoices("string", "str"))
    id: str | UUID = Field(default="")

    @field_validator("id")
    def id_validator(cls, id: str | UUID) -> str | UUID:
        if isinstance(id, UUID):
            return id
        try:
            uuid = UUID(id)
        except Exception:
            return id
        else:
            return uuid



from enum import IntEnum


# https://wiki.secondlife.com/wiki/Changed
class Change(IntEnum):
    INVENTORY = 0x001  #        Prim inventory has changed
    COLOR = 0x002  #            Prim color or alpha parameters have changed
    SHAPE = 0x004  #            Prim shape has changed
    SCALE = 0x008  #            Prim scale has changed
    TEXTURE = 0x010  #          Prim texture parameters have changed
    LINK = 0x020  #             The number of prims or avatars seated on the object have changed
    ALLOWED_DROP = 0x040  #     A user has added inventory to the prim
    OWNER = 0x080  #            The object has changed owners
    REGION = 0x100  #           The object has changed region
    TELEPORT = 0x200  #         The avatar has teleported
    REGION_START = 0x400  #     The region this object is in has just come online
    MEDIA = 0x800  #            Prim Media has changed
    RENDER_MATERIAL = 0x1000  # Render Material has changed



from pydantic import BaseModel, Field, field_validator

from lslgwlib.enums import PermValue


class Permission(BaseModel):
    perms: int = Field(ge=0, le=0x7FFFFFFF)

    def __init__(self, *args, **kwargs):
        if len(args) == 1:
            super().__init__(perms=args[0])
        else:
            super().__init__(**kwargs)

    @field_validator("perms")
    def perms_validator(cls, perms: int):
        if (
            perms & PermValue.COPY
            and perms & PermValue.MODIFY
            and perms & PermValue.MOVE
            and perms & PermValue.TRANSFER
        ):
            return PermValue.ALL
        match perms:
            case PermValue.COPY:
                return PermValue.COPY
            case PermValue.MODIFY:
                return PermValue.MODIFY
            case PermValue.MOVE:
                return PermValue.MOVE
            case PermValue.TRANSFER:
                return PermValue.TRANSFER
            case PermValue.NONE:
                return PermValue.NONE
        return perms

    @property
    def ALL(self) -> bool:
        return self.perms == PermValue.ALL

    @property
    def COPY(self) -> bool:
        return bool(self.perms & PermValue.COPY)

    @property
    def MODIFY(self) -> bool:
        return bool(self.perms & PermValue.MODIFY)

    @property
    def MOVE(self) -> bool:
        return bool(self.perms & PermValue.MOVE)

    @property
    def TRANSFER(self) -> bool:
        return bool(self.perms & PermValue.TRANSFER)


class Permissions(BaseModel):
    base: Permission
    owner: Permission
    group: Permission
    everyone: Permission
    next: Permission

    def __init__(self, *args, **kwargs) -> None:
        if len(args) == 5:
            super().__init__(
                base=Permission(args[0]),
                owner=Permission(args[1]),
                group=Permission(args[2]),
                everyone=Permission(args[3]),
                next=Permission(args[4]),
            )
        else:
            super().__init__(**kwargs)



from enum import IntEnum


# https://wiki.secondlife.com/wiki/LlGetInventoryType
class InvetoryType(IntEnum):
    ANY = -1  #       any inventory type
    TEXTURE = 0  #    texture
    SOUND = 1  #      sound
    LANDMARK = 3  #   landmark
    CLOTHING = 5  #   clothing
    OBJECT = 6  #     object
    NOTECARD = 7  #   notecard
    SCRIPT = 10  #    script
    BODYPART = 13  #  body part
    ANIMATION = 20  # animation
    GESTURE = 21  #   gesture
    SETTING = 56  #   setting
    MATERIAL = 57  #  material



from .permissions import PermMask, PermValue
from .attach_point import AttachPoint
from .chatchannel import ChatChannel
from .inventory import InvetoryType
from .linknum import LinkNum
from .change import Change


__all__ = [
    "AttachPoint",
    "Change",
    "LinkNum",
    "ChatChannel",
    "PermMask",
    "PermValue",
    "InvetoryType",
]



from pydantic import BaseModel, Field
from datetime import datetime
from uuid import UUID

from .avatar import Avatar
from .permissions import Permissions
from lslgwlib.enums import AttachPoint


class LinkSetInfo(BaseModel):
    id: UUID
    owner: Avatar
    lastOwnerId: UUID
    creatorId: UUID
    groupId: UUID
    name: str = Field(pattern=r"^[\x20-\x7b\x7d-\x7e]{1,63}$")
    description: str = Field(pattern=r"^[\x20-\x7b\x7d-\x7e]{0,127}$")
    attached: AttachPoint
    primsNum: int = Field(gt=0, le=255)
    inventoryNum: int = Field(ge=0, le=10000)
    createdAt: datetime
    rezzedAt: datetime
    scriptName: str = Field(pattern=r"^[\x20-\x7b\x7d-\x7e]{1,63}$")
    permissions: Permissions



from .inventory import Invetory, InvetoryItem
from .linksetinfo import LinkSetInfo
from .linkmessage import LinkMessage
from .chatmessage import ChatMessage
from .permissions import Permissions
from .priminfo import PrimInfo
from .httpdata import HTTPData
from .avatar import Avatar
from .region import Region
from .money import Money
from .touch import Touch


__all__ = [
    "LinkSetInfo",
    "HTTPData",
    "Avatar",
    "Region",
    "PrimInfo",
    "LinkMessage",
    "ChatMessage",
    "Money",
    "Touch",
    "Permissions",
    "Invetory",
    "InvetoryItem",
]



from enum import IntEnum


# https://wiki.secondlife.com/wiki/LlGetAttached
class AttachPoint(IntEnum):
    NONE = 0
    HEAD = 2  #           Skull                head
    NOSE = 17  #          Nose                 nose
    MOUTH = 11  #         Mouth                mouth
    FACE_TONGUE = 52  #   Tongue               tongue
    CHIN = 12  #          Chin                 chin
    FACE_JAW = 47  #      Jaw                  jaw
    LEAR = 13  #          Left Ear             left ear
    REAR = 14  #          Right Ear            right ear
    FACE_LEAR = 48  #     Alt Left Ear         left ear (extended)
    FACE_REAR = 49  #     Alt Right Ear        right ear (extended)
    LEYE = 15  #          Left Eye             left eye
    REYE = 16  #          Right Eye            right eye
    FACE_LEYE = 50  #     Alt Left Eye         left eye (extended)
    FACE_REYE = 51  #     Alt Right Eye        right eye (extended)
    NECK = 39  #          Neck                 neck
    LSHOULDER = 3  #      Left Shoulder        left shoulder
    RSHOULDER = 4  #      Right Shoulder       right shoulder
    LUARM = 20  #         L Upper Arm          left upper arm
    RUARM = 18  #         R Upper Arm          right upper arm
    LLARM = 21  #         L Lower Arm          left lower arm
    RLARM = 19  #         R Lower Arm          right lower arm
    LHAND = 5  #          Left Hand            left hand
    RHAND = 6  #          Right Hand           right hand
    LHAND_RING1 = 41  #   Left Ring Finger     left ring finger
    RHAND_RING1 = 42  #   Right Ring Finger    right ring finger
    LWING = 45  #         Left Wing            left wing
    RWING = 46  #         Right Wing           right wing
    CHEST = 1  #          Chest                chest/sternum
    LEFT_PEC = 29  #      Left Pec             left pectoral
    RIGHT_PEC = 30  #     Right Pec            right pectoral
    BELLY = 28  #         Stomach              belly/stomach/tummy
    BACK = 9  #           Spine                back
    TAIL_BASE = 43  #     Tail Base            tail base
    TAIL_TIP = 44  #      Tail Tip             tail tip
    AVATAR_CENTER = 40  # Avatar Center        avatar center/root
    PELVIS = 10  #        Pelvis               pelvis
    GROIN = 53  #         Groin                groin
    LHIP = 25  #          Left Hip             left hip
    RHIP = 22  #          Right Hip            right hip
    LULEG = 26  #         L Upper Leg          left upper leg
    RULEG = 23  #         R Upper Leg          right upper leg
    RLLEG = 24  #         R Lower Leg          right lower leg
    LLLEG = 27  #         L Lower Leg          left lower leg
    LFOOT = 7  #          Left Foot            left foot
    RFOOT = 8  #          Right Foot           right foot
    HIND_LFOOT = 54  #    Left Hind Foot       left hind foot
    HIND_RFOOT = 55  #    Right Hind Foot      right hind foot

    HUD_CENTER_2 = 31  #     HUD Center 2
    HUD_TOP_RIGHT = 32  #    HUD Top Right
    HUD_TOP_CENTER = 33  #   HUD Top
    HUD_TOP_LEFT = 34  #     HUD Top Left
    HUD_CENTER_1 = 35  #     HUD Center
    HUD_BOTTOM_LEFT = 36  #  HUD Bottom Left
    HUD_BOTTOM = 37  #       HUD Bottom
    HUD_BOTTOM_RIGHT = 38  # HUD Bottom Right



from enum import IntEnum


# https://wiki.secondlife.com/wiki/LlGetInventoryPermMask


class PermMask(IntEnum):
    BASE = 0  #     The base permissions
    OWNER = 1  #    Current owner permissions
    GROUP = 2  #    Active group permissions
    EVERYONE = 3  # Permissions everyone has
    NEXT = 4  #     Permissions the next owner will have


class PermValue(IntEnum):
    ALL = 0x7FFFFFFF  #      Move/Modify/Copy/Transfer permissions
    COPY = 0x00008000  #     Copy permission
    MODIFY = 0x00004000  #   Modify permission
    MOVE = 0x00080000  #     Move permission
    TRANSFER = 0x00002000  # Transfer permission
    NONE = 0  #              No one permission
